import functools
import time
from supervisely_lib.sly_logger import logger


def timeit(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        logger.debug(f"TIME {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer


import argparse
import io
import json
import requests

from supervisely_lib.io import fs as sly_fs
from supervisely_lib.nn.inference.rest_constants import GET_OUTPUT_META, IMAGE, INFERENCE, MODEL, SUPPORTED_REQUEST_TYPES

from requests_toolbelt import MultipartEncoder

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Inference REST client for standalone Supervisely models.')
    parser.add_argument('--server-url', required=True)
    parser.add_argument('--request-type', required=True, choices=SUPPORTED_REQUEST_TYPES)
    parser.add_argument('--in-image', default='')
    parser.add_argument('--out-json', default='')
    args = parser.parse_args()

    request_url = args.server_url + '/' + MODEL + '/' + args.request_type

    response_json = None
    if args.request_type == GET_OUTPUT_META:
        response = requests.post(request_url)
    elif args.request_type == INFERENCE:
        with open(args.in_image, 'rb') as fin:
            img_bytes = fin.read()
        img_ext = sly_fs.get_file_ext(args.in_image)
        encoder = MultipartEncoder({IMAGE: (args.in_image, io.BytesIO(img_bytes), 'application/octet-stream')})
        response = requests.post(request_url, data=encoder, headers={'Content-Type': encoder.content_type})
    else:
        raise ValueError(
            'Unknown model request type: {!r}. Only the following request types are supported: {!r}.'.format(
                args.request_type, SUPPORTED_REQUEST_TYPES))

    response.raise_for_status()
    response_str = json.dumps(response.json(), indent=4, sort_keys=True)

    if args.out_json:
        with open(args.out_json, 'w') as fout:
            fout.write(response_str)
    else:
        print(response_str)



# coding: utf-8
from supervisely_lib.geometry.bitmap import Bitmap
from supervisely_lib.geometry.cuboid import Cuboid
from supervisely_lib.geometry.point import Point
from supervisely_lib.geometry.polygon import Polygon
from supervisely_lib.geometry.polyline import Polyline
from supervisely_lib.geometry.rectangle import Rectangle
from supervisely_lib.geometry.graph import GraphNodes
from supervisely_lib.geometry.any_geometry import AnyGeometry
from supervisely_lib.geometry.cuboid_3d import Cuboid3d
from supervisely_lib.geometry.pointcloud import Pointcloud
from supervisely_lib.geometry.point_3d import Point3d
from supervisely_lib.geometry.multichannel_bitmap import MultichannelBitmap


_INPUT_GEOMETRIES = [Bitmap, Cuboid, Point, Polygon, Polyline, Rectangle, GraphNodes, AnyGeometry,
                     Cuboid3d, Pointcloud, Point3d, MultichannelBitmap]
_JSON_SHAPE_TO_GEOMETRY_TYPE = {geometry.geometry_name(): geometry for geometry in _INPUT_GEOMETRIES}


def GET_GEOMETRY_FROM_STR(figure_shape: str):
    '''
    The function create geometry class object from given string
    '''
    geometry = _JSON_SHAPE_TO_GEOMETRY_TYPE[figure_shape]
    return geometry



# coding: utf-8

from supervisely_lib.geometry.cuboid_3d import Vector3d
from supervisely_lib.geometry.geometry import Geometry
from supervisely_lib.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID


class Point3d(Geometry):
    @staticmethod
    def geometry_name():
        return 'point_3d'

    def __init__(self, point: Vector3d,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        if type(point) is not Vector3d:
            raise TypeError("\"position\" param has to be of type {!r}".format(type(Vector3d)))

        self._point = point

    @property
    def x(self):
        return self._point._x

    @property
    def y(self):
        return self._point._y

    @property
    def z(self):
        return self._point._z

    def to_json(self):
        res = self._point.to_json()
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        point = Vector3d.from_json(data)

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(point, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)


# coding: utf-8
import json

from supervisely_lib import Bitmap, logger, ObjClassCollection, ObjClass
from supervisely_lib.io import fs


BACKGROUND = 'background'
NEUTRAL = 'neutral'

CONTINUE_TRAINING = 'continue_training'
TRANSFER_LEARNING = 'transfer_learning'


def make_out_classes(in_project_classes, geometry_type, exclude_titles=None, extra_classes=None):
    exclude_titles = exclude_titles or []
    extra_classes = extra_classes or dict()

    if any(title in extra_classes for title in exclude_titles):
        raise RuntimeError(
            'Unable to construct class name to integer id map: extra classes names overlap with excluded classes.')

    out_classes = ObjClassCollection()
    for in_class in in_project_classes:
        if in_class.name not in exclude_titles:
            out_classes = out_classes.add(ObjClass(in_class.name, geometry_type, in_class.color))

    for title, color in extra_classes.items():
        if not out_classes.has_key(title):
            out_classes = out_classes.add(ObjClass(title, geometry_type, color))
    return out_classes


def create_segmentation_classes(in_project_classes, special_classes_config, bkg_input_idx,
                                weights_init_type, model_config_fpath, class_to_idx_config_key, start_class_id=1):
    extra_classes = {}
    special_class_ids = {}
    bkg_title = special_classes_config.get(BACKGROUND, None)
    if bkg_title is not None:
        extra_classes = {bkg_title: [34, 34, 34]} # Default background color
        special_class_ids[bkg_title] = bkg_input_idx

    exclude_titles = []
    neutral_title = special_classes_config.get(NEUTRAL, None)
    if neutral_title is not None:
        exclude_titles.append(neutral_title)
    out_classes = make_out_classes(in_project_classes, geometry_type=Bitmap, exclude_titles=exclude_titles,
                                   extra_classes=extra_classes)

    logger.info('Determined model out classes', extra={'out_classes': list(out_classes)})

    in_project_class_to_idx = make_new_class_to_idx_map(in_project_classes, start_class_id=start_class_id,
                                                        preset_class_ids=special_class_ids,
                                                        exclude_titles=exclude_titles)
    class_title_to_idx = infer_training_class_to_idx_map(weights_init_type,
                                                         in_project_class_to_idx,
                                                         model_config_fpath,
                                                         class_to_idx_config_key,
                                                         special_class_ids=special_class_ids)
    logger.info('Determined class mapping.', extra={'class_mapping': class_title_to_idx})
    return out_classes, class_title_to_idx


def make_new_class_to_idx_map(in_project_classes: ObjClassCollection, start_class_id: int,
                              preset_class_ids: dict = None, exclude_titles: list = None):
    preset_class_ids = preset_class_ids or dict()
    exclude_titles = exclude_titles or []
    if any(title in preset_class_ids for title in exclude_titles):
        raise RuntimeError(
            'Unable to construct class name to integer id map: preset classes names overlap with excluded classes.')
    sorted_titles = sorted([obj_class.name for obj_class in in_project_classes])
    if len(sorted_titles) != len(set(sorted_titles)):
        raise RuntimeError('Unable to construct class name to integer id map: class names are not unique.')

    class_title_to_idx = preset_class_ids.copy()
    next_title_id = start_class_id
    already_used_ids = set(class_title_to_idx.values())
    for title in sorted_titles:
        if title not in class_title_to_idx and title not in exclude_titles:
            while next_title_id in already_used_ids:
                next_title_id += 1
            class_title_to_idx[title] = next_title_id
            already_used_ids.add(next_title_id)
    return class_title_to_idx


def read_validate_model_class_to_idx_map(model_config_fpath, in_project_classes_set, class_to_idx_config_key):
    """Reads class id --> int index mapping from the model config; checks that the set of classes matches the input."""

    if not fs.file_exists(model_config_fpath):
        raise RuntimeError('Unable to continue_training, config for previous training wasn\'t found.')

    with open(model_config_fpath) as fin:
        model_config = json.load(fin)

    model_class_mapping = model_config.get(class_to_idx_config_key, None)
    if model_class_mapping is None:
        raise RuntimeError('Unable to continue_training, model does not have class mapping information.')
    model_classes_set = set(model_class_mapping.keys())

    if model_classes_set != in_project_classes_set:
        error_message_text = 'Unable to continue_training, sets of classes for model and dataset do not match.'
        logger.critical(
            error_message_text, extra={'model_classes': model_classes_set, 'dataset_classes': in_project_classes_set})
        raise RuntimeError(error_message_text)
    return model_class_mapping.copy()


def infer_training_class_to_idx_map(weights_init_type, in_project_class_to_idx, model_config_fpath,
                                    class_to_idx_config_key, special_class_ids=None):
    if weights_init_type == TRANSFER_LEARNING:
        logger.info('Transfer learning mode, using a class mapping created from scratch.')
        class_title_to_idx = in_project_class_to_idx
    elif weights_init_type == CONTINUE_TRAINING:
        logger.info('Continued training mode, reusing the existing class mapping from the model.')
        class_title_to_idx = read_validate_model_class_to_idx_map(
            model_config_fpath=model_config_fpath,
            in_project_classes_set=set(in_project_class_to_idx.keys()),
            class_to_idx_config_key=class_to_idx_config_key)
    else:
        raise RuntimeError('Unknown weights init type: {}'.format(weights_init_type))

    if special_class_ids is not None:
        for class_title, requested_class_id in special_class_ids.items():
            effective_class_id = class_title_to_idx[class_title]
            if requested_class_id != effective_class_id:
                error_msg = ('Unable to start training. Effective integer id for class {} does not match the ' +
                             'requested value in the training config ({} vs {}).'.format(
                                 class_title, effective_class_id, requested_class_id))
                logger.critical(error_msg, extra={'class_title_to_idx': class_title_to_idx,
                                                  'special_class_ids': special_class_ids})
                raise RuntimeError(error_msg)
    return class_title_to_idx



# coding: utf-8
from supervisely_lib.collection.str_enum import StrEnum


# ['admin',  'developer', 'manager', 'reviewer', 'annotator', 'viewer']
class UserRoleName(StrEnum):
    ADMIN = 'admin'
    DEVELOPER = 'developer'
    MANAGER = 'manager'
    REVIEWER = 'reviewer'
    ANNOTATOR = 'annotator'
    VIEWER = 'viewer'



# coding: utf-8

from copy import deepcopy
import uuid

from supervisely_lib._utils import take_with_default
from supervisely_lib.video_annotation.video_tag_collection import VideoTagCollection
from supervisely_lib.video_annotation.video_object_collection import VideoObjectCollection
from supervisely_lib.video_annotation.frame_collection import FrameCollection
from supervisely_lib.video_annotation.constants import FRAMES, IMG_SIZE, IMG_SIZE_HEIGHT, IMG_SIZE_WIDTH, \
                                                       DESCRIPTION, FRAMES_COUNT, TAGS, OBJECTS, VIDEO_ID, KEY, \
                                                       VIDEOS_MAP, VIDEO_NAME
from supervisely_lib.video_annotation.key_id_map import KeyIdMap


class VideoAnnotation:
    '''
    This is a class for creating and using annotations for videos
    '''
    def __init__(self, img_size, frames_count, objects=None, frames=None, tags=None, description="", key=None):
        '''
        The constructor for VideoAnnotation class.
        :param img_size: size of the image(tuple or list of integers)
        :param frames_count: int
        :param objects: VideoObjectCollection
        :param frames: FrameCollection
        :param tags: VideoTagCollection
        :param description: str
        :param key: uuid class object
        '''
        if not isinstance(img_size, (tuple, list)):
            raise TypeError('{!r} has to be a tuple or a list. Given type "{}".'.format('img_size', type(img_size)))
        self._img_size = tuple(img_size)
        self._frames_count = frames_count

        self._description = description
        self._tags = take_with_default(tags, VideoTagCollection())
        self._objects = take_with_default(objects, VideoObjectCollection())
        self._frames = take_with_default(frames, FrameCollection())
        self._key = take_with_default(key, uuid.uuid4())

        self.validate_figures_bounds()

    @property
    def img_size(self):
        return deepcopy(self._img_size)

    @property
    def frames_count(self):
        return self._frames_count

    @property
    def objects(self):
        return self._objects

    @property
    def frames(self):
        return self._frames

    @property
    def figures(self):
        '''
        :return: list of figures from all frames in collection
        '''
        return self.frames.figures

    @property
    def tags(self):
        return self._tags

    def key(self):
        return self._key

    @property
    def description(self):
        return self._description

    def validate_figures_bounds(self):
        '''
        The function validate_figures_bounds checks if image contains figures from all frames in collection. Raise error if figure is out of image bounds
        '''
        for frame in self.frames:
            frame.validate_figures_bounds(self.img_size)

    def to_json(self, key_id_map: KeyIdMap=None):
        '''
        The function to_json convert videoannotation to json format
        :param key_id_map: KeyIdMap class object
        :return: videoannotation in json format
        '''
        res_json = {
                        IMG_SIZE: {
                                    IMG_SIZE_HEIGHT: int(self.img_size[0]),
                                    IMG_SIZE_WIDTH: int(self.img_size[1])
                                  },
                        DESCRIPTION: self.description,
                        KEY: self.key().hex,
                        TAGS: self.tags.to_json(key_id_map),
                        OBJECTS: self.objects.to_json(key_id_map),
                        FRAMES: self.frames.to_json(key_id_map),
                        FRAMES_COUNT: self.frames_count
                    }

        if key_id_map is not None:
            video_id = key_id_map.get_video_id(self.key())
            if video_id is not None:
                res_json[VIDEO_ID] = video_id

        return res_json

    @classmethod
    def from_json(cls, data, project_meta, key_id_map: KeyIdMap=None):
        '''
        The function from_json convert videoannotation from json format to VideoAnnotation class object.
        :param data: input videoannotation in json format
        :param project_meta: ProjectMeta class object
        :param key_id_map: KeyIdMap class object
        :return: VideoAnnotation class object
        '''
        #video_name = data[VIDEO_NAME]
        video_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(video_key, data.get(VIDEO_ID, None))

        img_size_dict = data[IMG_SIZE]
        img_height = img_size_dict[IMG_SIZE_HEIGHT]
        img_width = img_size_dict[IMG_SIZE_WIDTH]
        img_size = (img_height, img_width)

        description = data.get(DESCRIPTION, "")
        frames_count = data[FRAMES_COUNT]

        tags = VideoTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = VideoObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)
        frames = FrameCollection.from_json(data[FRAMES], objects, frames_count, key_id_map)

        return cls(img_size=img_size,
                   frames_count=frames_count,
                   objects=objects,
                   frames=frames,
                   tags=tags,
                   description=description,
                   key=video_key)

    def clone(self, img_size=None, frames_count=None, objects=None, frames=None, tags=None, description=None):
        '''
        :param img_size: size of the image(tuple or list of integers)
        :param frames_count: int
        :param objects: VideoObjectCollection
        :param frames: FrameCollection
        :param tags: VideoTagCollection
        :param description: str
        :return: VideoAnnotation class object
        '''
        return VideoAnnotation(img_size=take_with_default(img_size, self.img_size),
                               frames_count=take_with_default(frames_count, self.frames_count),
                               objects=take_with_default(objects, self.objects),
                               frames=take_with_default(frames, self.frames),
                               tags=take_with_default(tags, self.tags),
                               description=take_with_default(description, self.description))


    def is_empty(self):
        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False



# coding: utf-8

import uuid
from bidict import bidict
from supervisely_lib.io.json import dump_json_file, load_json_file

TAGS = 'tags'
OBJECTS = 'objects'
FIGURES = 'figures'
VIDEOS = 'videos'

ALLOWED_KEY_TYPES = [TAGS, OBJECTS, VIDEOS, FIGURES]


class KeyIdMap:
    '''
    This is a class for creating and using key_id_maps for videos
    '''
    def __init__(self):
        self._data = dict()
        self._data[TAGS] = bidict()
        self._data[OBJECTS] = bidict()
        self._data[FIGURES] = bidict()
        self._data[VIDEOS] = bidict()

    def _add(self, key_type, key: uuid.UUID, id: int = None):
        '''
        Add given data in self._data dictionary. Raise error if data type of any parameter is invalid
        :param key_type: str
        :param key: uuid class object
        :param id: int
        '''
        if key_type not in ALLOWED_KEY_TYPES:
            raise RuntimeError("Key type {!r} is not allowed. Allowed types are {}".format(key_type, ALLOWED_KEY_TYPES))
        if type(key) is not uuid.UUID:
            raise RuntimeError("Key should be of type uuid.UUID")
        if id is not None and type(id) is not int:
            raise RuntimeError("Id should be of type int")
        self._data[key_type].update(bidict({key: id}))

    def add_object(self, key: uuid.UUID, id: int):
        '''
        Add objects in self._data dictionary with given key and id.
        :param key: uuid class object
        :param id: int
        '''
        self._add(OBJECTS, key, id)

    def add_tag(self, key: uuid.UUID, id: int):
        '''
        Add tags in self._data dictionary with given key and id.
        :param key: uuid class object
        :param id: int
        '''
        self._add(TAGS, key, id)

    def add_figure(self, key: uuid.UUID, id: int):
        '''
        Add figures in self._data dictionary with given key and id.
        :param key: uuid class object
        :param id: int
        '''
        self._add(FIGURES, key, id)

    def add_video(self, key: uuid.UUID, id: int):
        '''
        Add videos in self._data dictionary with given key and id.
        :param key: uuid class object
        :param id: int
        '''
        self._add(VIDEOS, key, id)

    def _get_id_by_key(self, key_type, key: uuid.UUID):
        '''
        :param key_type: str
        :param key: uuid class object
        :return: Id by given key. None if there is no such key. Raise error if key type is not uuid.UUID
        '''
        if type(key) is not uuid.UUID:
            raise RuntimeError("Key should be of type uuid.UUID")

        if key in self._data[key_type]:
            return self._data[key_type][key]
        else:
            return None

    def _get_key_by_id(self, key_type, id: int):
        '''
        :param key_type: str
        :param id: int
        :return: Key by given id. None if there is no such id. Raise error if id type is not int
        '''
        if type(id) is not int:
            raise RuntimeError("Id should be of type int")
        if id not in self._data[key_type].inverse:
            return None
        return self._data[key_type].inverse[id]

    def get_object_id(self, key: uuid.UUID):
        '''
        :param key: uuid class object
        :return: Id by given key for objects field.
        '''
        return self._get_id_by_key(OBJECTS, key)

    def get_tag_id(self, key: uuid.UUID):
        '''
        :param key: uuid class object
        :return: Id by given key for tags field.
        '''
        return self._get_id_by_key(TAGS, key)

    def get_figure_id(self, key: uuid.UUID):
        '''
        :param key: uuid class object
        :return: Id by given key for figures field.
        '''
        return self._get_id_by_key(FIGURES, key)

    def get_video_id(self, key: uuid.UUID):
        '''
        :param key: uuid class object
        :return: Id by given key for videos field.
        '''
        return self._get_id_by_key(VIDEOS, key)

    def get_object_key(self, id: int):
        '''
        :param id: int
        :return: Key by given id for objects field.
        '''
        return self._get_key_by_id(OBJECTS, id)

    def get_tag_key(self, id: int):
        '''
        :param id: int
        :return: Key by given id for tags field.
        '''
        return self._get_key_by_id(TAGS, id)

    def get_figure_key(self, id: int):
        '''
        :param id: int
        :return: Key by given id for figures field.
        '''
        return self._get_key_by_id(FIGURES, id)

    def get_video_key(self, id: int):
        '''
        :param id: int
        :return: Key by given id for videos field.
        '''
        return self._get_key_by_id(VIDEOS, id)

    def to_dict(self):
        '''
        Convert current self._data dictionary with bidict values to dictionary with dict values
        :return: dict
        '''
        simple_dict = {}
        for type_str, value_bidict in self._data.items():
            sub_dict = {}
            for uuid, int_id in value_bidict.items():
                sub_dict[uuid.hex] = int_id
            simple_dict[type_str] = sub_dict
        return simple_dict

    def dump_json(self, path):
        '''
        Save current class object data in json format by given path
        :param path: str
        '''
        simple_dict = self.to_dict()
        dump_json_file(simple_dict, path, indent=4)

    @classmethod
    def load_json(cls, path):
        '''
        Download json data by given path and convert in to dict with bidict values
        :param path: str
        :return: dict
        '''
        simple_dict = load_json_file(path)
        result = cls()
        for key_type, value_dict in simple_dict.items():
            for key_str, id in value_dict.items():
                result._add(key_type, uuid.UUID(key_str), id)
        return result

    @classmethod
    def _add_to(cls, key_id_map, key_type, keys, ids):
        '''
        Add given values(keys, ids) to KeyIdMap class object with given type of key
        :param key_id_map: KeyIdMap class object
        :param key_type: str
        :param keys: list of uuid class objects
        :param ids: list of integers
        :return: None if key_id_map parameter is None
        '''
        if key_id_map is None:
            return
        for key, id in zip(keys, ids):
            key_id_map._add(key_type, key, id)

    @classmethod
    def add_tags_to(cls, key_id_map, keys, ids):
        '''
        Add tags type of key with given values(keys, ids) to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param keys: list of uuid class objects
        :param ids: list of integers
        '''
        cls._add_to(key_id_map, TAGS, keys, ids)

    @classmethod
    def add_tag_to(cls, key_id_map, key, id):
        '''
        Add tags type of key with given key and id to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param key: uuid class object
        :param id: int
        '''
        cls._add_tags_to(key_id_map, [key], [id])

    @classmethod
    def add_objects_to(cls, key_id_map, keys, ids):
        '''
        Add objects type of key with given values(keys, ids) to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param keys: list of uuid class objects
        :param ids: list of integers
        '''
        cls._add_to(key_id_map, OBJECTS, keys, ids)

    @classmethod
    def add_object_to(cls, key_id_map, key, id):
        '''
        Add objects type of key with given key and id to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param key: uuid class object
        :param id: int
        '''
        cls._add_objects_to(key_id_map, [key], [id])

    @classmethod
    def add_figures_to(cls, key_id_map, keys, ids):
        '''
        Add figures type of key with given values(keys, ids) to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param keys: list of uuid class objects
        :param ids: list of integers
        '''
        cls._add_to(key_id_map, FIGURES, keys, ids)

    @classmethod
    def add_figure_to(cls, key_id_map, key, id):
        '''
        Add figure type of key with given key and id to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param key: uuid class object
        :param id: int
        '''
        cls._add_figures_to(key_id_map, [key], [id])

    @classmethod
    def add_videos_to(cls, key_id_map, keys, ids):
        '''
        Add videos type of key with given values(keys, ids) to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param keys: list of uuid class objects
        :param ids: list of integers
        '''
        cls._add_to(key_id_map, VIDEOS, keys, ids)

    @classmethod
    def add_video_to(cls, key_id_map, key, id):
        '''
        Add video type of key with given key and id to KeyIdMap class object
        :param key_id_map: KeyIdMap class object
        :param key: uuid class object
        :param id: int
        '''
        cls._add_videos_to(key_id_map, [key], [id])



# coding: utf-8

import uuid

from supervisely_lib._utils import take_with_default
from supervisely_lib.api.module_api import ApiField
from supervisely_lib.pointcloud_annotation.pointcloud_object_collection import PointcloudObjectCollection
from supervisely_lib.video_annotation.constants import FRAMES, DESCRIPTION, FRAMES_COUNT, TAGS, OBJECTS, KEY
from supervisely_lib.video_annotation.frame_collection import FrameCollection
from supervisely_lib.video_annotation.key_id_map import KeyIdMap
from supervisely_lib.video_annotation.video_tag_collection import VideoTagCollection


class PointcloudEpisodeAnnotation:
    def __init__(self, frames_count=None, objects=None, frames=None, tags=None, description="", key=None):
        self._frames_count = frames_count
        self._description = description
        self._frames = take_with_default(frames, FrameCollection())
        self._tags = take_with_default(tags, VideoTagCollection())
        self._objects = take_with_default(objects, PointcloudObjectCollection())
        self._key = take_with_default(key, uuid.uuid4())

    def to_json(self, key_id_map: KeyIdMap = None):
        res_json = {
            DESCRIPTION: self.description,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            FRAMES_COUNT: self.frames_count,
            FRAMES: self.frames.to_json(key_id_map),
        }

        if key_id_map is not None:
            dataset_id = key_id_map.get_video_id(self.key())
            if dataset_id is not None:
                res_json[ApiField.DATASET_ID] = dataset_id

        return res_json

    @classmethod
    def from_json(cls, data, project_meta, key_id_map: KeyIdMap = None):
        item_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(item_key, data.get(ApiField.DATASET_ID, None))

        description = data.get(DESCRIPTION, "")
        frames_count = data.get(FRAMES_COUNT, 0)

        tags = VideoTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = PointcloudObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)
        frames = FrameCollection.from_json(data[FRAMES], objects, key_id_map=key_id_map)

        return cls(frames_count, objects, frames, tags, description, item_key)

    def clone(self, frames_count=None, objects=None, frames=None, tags=None, description=""):
        return PointcloudEpisodeAnnotation(frames_count=take_with_default(frames_count, self.frames_count),
                                           objects=take_with_default(objects, self.objects),
                                           frames=take_with_default(frames, self.frames),
                                           tags=take_with_default(tags, self.tags),
                                           description=take_with_default(description, self.description))

    @property
    def frames_count(self):
        return self._frames_count

    @property
    def objects(self):
        return self._objects

    @property
    def frames(self):
        return self._frames

    @property
    def figures(self):
        return self.frames.figures

    @property
    def tags(self):
        return self._tags

    def key(self):
        return self._key

    @property
    def description(self):
        return self._description

    def is_empty(self):
        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False



# coding: utf-8

import os
import queue
import time
from collections import namedtuple
from copy import deepcopy
from threading import Thread

import multiprocessing as mp

from supervisely_lib import logger
from supervisely_lib.annotation.annotation import Annotation
from supervisely_lib.imaging import image as sly_image
from supervisely_lib.nn.config import AlwaysPassingConfigValidator
from supervisely_lib.project.project import Project, read_single_project, OpenMode
from supervisely_lib.project.project_meta import ProjectMeta
from supervisely_lib.task.paths import TaskPaths
from supervisely_lib.task.progress import report_inference_finished
from supervisely_lib.nn.hosted.inference_modes import InferenceModeFactory
from supervisely_lib.nn.hosted.inference_batch import determine_task_inference_mode_config
from supervisely_lib.task.progress import Progress


InferenceRequest = namedtuple('InferenceRequest', ['ds_name', 'item_name', 'item_paths'])
InferenceResponse = namedtuple('InferenceResponse', ['ds_name', 'item_name', 'item_paths', 'ann_json', 'meta_json'])


def single_inference_process_fn(inference_initializer, inference_mode_config, in_project_meta_json, request_queue,
                                result_meta_queue, progress_queue, project):
    """Loads a separate model, processes requests from request_queue, results go to result_queue.

    None request signals the process to finish.
    """
    single_image_inference = inference_initializer()
    inference_mode = InferenceModeFactory.create(
        inference_mode_config, ProjectMeta.from_json(in_project_meta_json), single_image_inference)

    project_meta_sent = False
    req = ''
    while req is not None:
        req = request_queue.get()
        if req is not None:
            # Send the resulting project meta to the parent project to make sure we only write the meta JSON once.
            if not project_meta_sent:
                try:
                    result_meta_queue.put(inference_mode.out_meta.to_json(), block=False)
                except queue.Full:
                    pass
            project_meta_sent = True

            in_ann = Annotation.load_json_file(req.item_paths.ann_path, inference_mode.out_meta)
            ann = inference_mode.infer_annotate_image_file(req.item_paths.img_path, in_ann)
            out_dataset = project.datasets.get(req.ds_name)
            out_dataset.add_item_file(
                req.item_name, req.item_paths.img_path, ann=ann, _validate_item=False, _use_hardlink=True)
            progress_queue.put(1)


def progress_report_thread_fn(in_project, progress_queue):
    """Gets inference result annotations from the queue and writes them to the output dataset.
    None result signals the thread to finish.
    """

    progress_bar = Progress('Model applying: ', in_project.total_items)
    while True:
        resp = progress_queue.get()
        if resp is not None:
            progress_bar.iter_done_report()
        else:
            break


def populate_inference_requests_queue(in_project, inference_processes, request_queue):
    for in_dataset in in_project:
        for in_item_name in in_dataset:
            logger.trace('Will process image',
                         extra={'dataset_name': in_dataset.name, 'image_name': in_item_name})
            in_item_paths = in_dataset.get_item_paths(in_item_name)
            req = InferenceRequest(ds_name=in_dataset.name, item_name=in_item_name,
                                   item_paths=in_item_paths)
            while True:
                try:
                    # Set a finite timeout for the queue insertion attempt to prevent deadlocks if all of the
                    # inference processes die in the interim.
                    request_queue.put(req, timeout=0.1)
                    break
                except queue.Full:
                    # If any of the inference processes has died, stop populating the requests queue and exit right
                    # away. Otherwise a deadlock is possible if no inference processes survive to take requests off
                    # the queue.
                    if not all(p.is_alive() for p in inference_processes):
                        # Early exit, return False to indicate failure.
                        return False
    return True


class BatchInferenceMultiprocessApplier:
    QUEUE_ELEMENTS_PER_PROCESS = 40

    def __init__(self, single_image_inference_initializer, num_processes, default_inference_mode_config: dict,
                 config_validator=None):
        self._config_validator = config_validator or AlwaysPassingConfigValidator()
        self._inference_mode_config = determine_task_inference_mode_config(deepcopy(default_inference_mode_config))

        self._in_project = read_single_project(TaskPaths.DATA_DIR)
        logger.info('Project structure has been read. Samples: {}.'.format(self._in_project.total_items))

        out_dir = os.path.join(TaskPaths.RESULTS_DIR, self._in_project.name)
        self._out_project = Project(out_dir, OpenMode.CREATE)
        # Create the output datasets in advance so that the worker processes do not collide on creating their own on
        # disk.
        for in_dataset in self._in_project:
            self._out_project.create_dataset(in_dataset.name)

        self._inference_request_queue = mp.Queue(maxsize=self.QUEUE_ELEMENTS_PER_PROCESS * num_processes)
        self._result_meta_queue = mp.Queue(maxsize=3)
        self._progress_report_queue = mp.Queue(maxsize=self.QUEUE_ELEMENTS_PER_PROCESS * num_processes)
        self._inference_processes = [
            mp.Process(
                target=single_inference_process_fn,
                args=(single_image_inference_initializer, self._inference_mode_config, self._in_project.meta.to_json(),
                      self._inference_request_queue, self._result_meta_queue,
                      self._progress_report_queue, self._out_project),
                daemon=True)
            for _ in range(num_processes)]
        logger.info('Dataset inference preparation done.')
        for p in self._inference_processes:
            p.start()

    def run_inference(self):
        progress_report_thread = Thread(
            target=progress_report_thread_fn, args=(self._in_project, self._progress_report_queue), daemon=True)
        progress_report_thread.start()

        feed_status = populate_inference_requests_queue(
            self._in_project, self._inference_processes, self._inference_request_queue)
        for _ in self._inference_processes:
            self._inference_request_queue.put(None)

        out_meta_json = self._result_meta_queue.get()
        self._out_project.set_meta(ProjectMeta.from_json(out_meta_json))

        for p in self._inference_processes:
            p.join()

        if not feed_status or not all(p.exitcode == 0 for p in self._inference_processes):
            raise RuntimeError('One of the inference processes encountered an error.')

        self._progress_report_queue.put(None)
        progress_report_thread.join()
        report_inference_finished()
